import math

import cv2 as cv

from gamedriver._geometry import Box, tap_box
from gamedriver._locate import locate
from gamedriver._util import wait
from gamedriver.logger import logger
from gamedriver.settings import settings


def wait_until_img_visible(
    img: str,
    *,
    bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
    timeout_s=30,
) -> Box | None:
    """Waits until an image is visible.

    Polls to check whether the image is visible every
    :py:attr:`settings.Settings.refresh_rate_ms` milliseconds.

    See :py:func:`gamedriver.locate`.

    Args:
        timeout_s (int, optional): Timeout in seconds. Defaults to 30.

    Returns:
        Box | None: The match, or None if there is no match within `timeout_s`.
    """
    polling_interval_s = settings["refresh_rate_ms"] / 1_000

    box = None
    for i in range(math.floor(timeout_s / polling_interval_s)):
        box = locate(
            img,
            bounding_box=bounding_box,
            convert_to_grayscale=convert_to_grayscale,
            is_grayscale=is_grayscale,
            method=method,
            threshold=threshold,
        )
        if box:
            logger.debug(f"{img} available after {i * polling_interval_s}s")
            break
        wait(polling_interval_s)
    else:
        logger.debug(f"{img} not available after {timeout_s}s")

    return box


def tap_img(
    img: str,
    *,
    bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
) -> Box | None:
    """Taps the center of an image if found.

    See :py:func:`gamedriver.locate`.
    """
    box = locate(
        img,
        bounding_box=bounding_box,
        convert_to_grayscale=convert_to_grayscale,
        is_grayscale=is_grayscale,
        method=method,
        threshold=threshold,
    )
    if not box:
        logger.debug(
            f"{img} not found{f' in bounding box {bounding_box}' if bounding_box else ''}"
        )
    else:
        tap_box(box)
    return box


# Makes us seem a little more human, if you're into that ;) (at the expense of speed)
def tap_img_when_visible_after_wait(
    img: str,
    *,
    bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
    timeout_s=30,
    seconds=1,
) -> Box | None:
    """Waits until an image is visible, waits for a time, then taps it.

    See :py:func:`gamedriver.locate` and
    :py:func:`gamedriver.wait_until_img_visible`.

    Args:
        seconds (int, optional): Time to wait after the image becomes visible
            before tapping it. Defaults to 1.
    """
    box = wait_until_img_visible(
        img,
        bounding_box=bounding_box,
        convert_to_grayscale=convert_to_grayscale,
        is_grayscale=is_grayscale,
        method=method,
        threshold=threshold,
        timeout_s=timeout_s,
    )
    if box:
        wait(seconds)
        tap_box(box)
    return box


def tap_img_when_visible(
    img: str,
    *,
    bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
    timeout_s=30,
) -> Box | None:
    """Waits until an image is visible, then taps the center of it.

    See :py:func:`gamedriver.locate` and
    :py:func:`gamedriver.wait_until_img_visible`.
    """
    return tap_img_when_visible_after_wait(
        img,
        bounding_box=bounding_box,
        convert_to_grayscale=convert_to_grayscale,
        is_grayscale=is_grayscale,
        method=method,
        threshold=threshold,
        timeout_s=timeout_s,
        seconds=0,
    )


def tap_img_while_other_visible(
    img: str,
    other: str,
    *,
    bounding_box: Box = None,
    other_bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
    timeout_s=5,
    frequency_s=1,
) -> bool:
    """Taps an image while another is visible.

    See :py:func:`gamedriver.locate` and
    :py:func:`gamedriver.wait_until_img_visible`.

    Args:
        img (str): Image to tap
        other (str): Image to check the visibility of
        frequency_s (int, optional): How often the image to tap should be
            tapped, in seconds. Defaults to 1.

    Returns:
        bool: True if the image to check disappeared within the timeout and
            tapping was stopped, or False if the image to check remained after
            the timeout and tapping was continued until the end.
    """
    tap_count = math.floor(timeout_s / frequency_s)
    for _ in range(tap_count):
        if not locate(
            other,
            bounding_box=other_bounding_box,
            convert_to_grayscale=convert_to_grayscale,
            is_grayscale=is_grayscale,
            method=method,
            threshold=threshold,
        ):
            break
        tap_img(
            img,
            bounding_box=bounding_box,
            convert_to_grayscale=convert_to_grayscale,
            is_grayscale=is_grayscale,
            method=method,
            threshold=threshold,
        )
        wait(frequency_s)
    else:
        logger.error(
            f"Kept tapping image {img}, but image {other} was still visible "
            + f"after {tap_count} tries each {frequency_s} seconds apart"
        )
        return False

    return True


def tap_img_while_visible(
    img: str,
    *,
    bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
    timeout_s=5,
    frequency_s=1,
) -> bool:
    """Taps an image while it is visible.

    See :py:func:`tap_img_while_other_visible`.

    Args:
        img (str): Image to check the visibility of and tap
    """
    # Touching an image while it is visible is a special case of taping it
    # while an arbitrary image is visible
    return tap_img_while_other_visible(
        img,
        img,
        bounding_box=bounding_box,
        other_bounding_box=bounding_box,
        convert_to_grayscale=convert_to_grayscale,
        is_grayscale=is_grayscale,
        method=method,
        threshold=threshold,
        timeout_s=timeout_s,
        frequency_s=frequency_s,
    )



from collections.abc import Callable
from typing import Optional, TypedDict

import cv2 as cv


class Settings(TypedDict):
    """GameDriver configuration."""

    #: How much to multiply :py:func:`gamedriver.wait` s by. Defaults to 1.
    wait_scale: float
    #: How much to add to :py:func:`gamedriver.wait` s. Defaults to 0.
    wait_offset: float

    #: Path to the template image directory. Recommend using an absolute path.
    #: Defaults to "".
    img_path: str
    #: File extension for template images. Defaults to ".png".
    img_ext: str

    #: Returns the screen [height, width, channel] in the BGR color space.
    #: Required for template matching that uses the device screen.
    get_screen: Callable[[], cv.typing.MatLike]
    #: How often the screen is checked in explicit wait operations. For example
    #: when using :py:func:`gamedriver.wait_until_img_visible`. Defaults to
    #: 1000 (1s).
    refresh_rate_ms: int

    #: Taps an (`x`, `y`) coordinate. Required for all tap-based interactions.
    tap_xy: Callable[[int, int], None]
    #: Swipes from (`x1`, `y1`) to (`x2`, `y2`) in `duration_ms` milliseconds.
    #: Required for all swipe interactions.
    swipe: Callable[[int, int, int, int, Optional[int]], None]


def _raise_default_setting_err(prop: str) -> None:
    raise ValueError(f"{prop} setting not set")


default_settings: Settings = {
    "wait_scale": 1,
    "wait_offset": 0,
    "img_path": "",
    "img_ext": ".png",
    "get_screen": lambda: _raise_default_setting_err("get_screen"),
    "refresh_rate_ms": 1_000,
    "tap_xy": lambda *args: _raise_default_setting_err("tap_xy"),
    "swipe": lambda *args: _raise_default_setting_err("swipe"),
}


settings: Settings = default_settings


def set_settings(s: Settings) -> None:
    """Sets multiple settings.

    A convenient way to set multiple settings at once. Only settings set in
    this way will be affected - existing settings that are not set remain
    unaffected.

    Args:
        s (Settings): Partial settings dictionary with the settings to be set
    """
    for k, v in s.items():
        settings[k] = v



from gamedriver._geometry import Box, get_center, Point, tap_box
from gamedriver._locate import locate, locate_all
from gamedriver._match_template import match_template, match_template_all
from gamedriver._tap_img import (
    tap_img,
    tap_img_when_visible,
    tap_img_when_visible_after_wait,
    tap_img_while_other_visible,
    tap_img_while_visible,
    wait_until_img_visible,
)
from gamedriver._util import get_img_path, get_pixel, get_screen, swipe, tap_xy, wait

from gamedriver.logger import _setup_logging


_setup_logging()



from typing import Iterator

import cv2 as cv

from gamedriver._geometry import Box
from gamedriver._match_template import match_template, match_template_all
from gamedriver._util import get_img_path, get_screen


def locate(
    img: str,
    *,
    bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
) -> Box | None:
    """Locates an image on the screen.

    Performs a template match where the haystack image is the device screen and
    the needle image is specified by the relative path instead of the full path.

    See :py:func:`gamedriver.match_template`.

    See also :py:func:`gamedriver.get_screen` and
    :py:func:`gamedriver.get_img_path`.

    Args:
        img (str): Relative path to the needle image, with no file extension

    Returns:
        Box | None

    Examples:
        ::

            box = gd.locate("buttons/confirm")
    """
    screen = (
        get_screen(grayscale=is_grayscale) if is_grayscale is not None else get_screen()
    )
    return match_template(
        screen,
        get_img_path(img),
        bounding_box=bounding_box,
        convert_to_grayscale=convert_to_grayscale,
        is_grayscale=is_grayscale,
        method=method,
        threshold=threshold,
    )


def locate_all(
    img: str,
    *,
    bounding_box: Box = None,
    convert_to_grayscale=True,
    is_grayscale=False,
    method=cv.TM_SQDIFF_NORMED,
    threshold: float = None,
) -> Iterator[Box]:
    """Locates all images on the screen.

    See :py:func:`locate`.

    See also :py:func:`gamedriver.match_template_all`.

    Yields:
        Iterator[Box]
    """
    screen = (
        get_screen(grayscale=is_grayscale) if is_grayscale is not None else get_screen()
    )
    return match_template_all(
        screen,
        get_img_path(img),
        bounding_box=bounding_box,
        convert_to_grayscale=convert_to_grayscale,
        is_grayscale=is_grayscale,
        method=method,
        threshold=threshold,
    )



import atexit
import logging
import logging.handlers
import queue
import sys

from pythonjsonlogger import jsonlogger


logger = logging.getLogger("gamedriver")


def _setup_logging() -> None:
    formatter = jsonlogger.JsonFormatter("%(levelname)s %(message)s")
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setFormatter(formatter)
    # TODO: Get log path and enable file logging
    # file_handler = logging.handlers.RotatingFileHandler(
    #     "log.jsonl", maxBytes=10 * 1024 * 1024, backupCount=2
    # )
    # file_handler.setFormatter(formatter)

    # # Use a queue for the non-critical file logging so it doesn't block.
    # # See https://docs.python.org/3/howto/logging-cookbook.html#dealing-with-handlers-that-block
    # q = queue.Queue(-1)  # no limit on size
    # queue_handler = logging.handlers.QueueHandler(q)
    # queue_listener = logging.handlers.QueueListener(
    #     q, file_handler, respect_handler_level=True
    # )
    # queue_listener.start()
    # atexit.register(queue_listener.stop)

    logging.basicConfig(
        # encoding="utf-8", handlers=[stdout_handler, queue_handler]
        encoding="utf-8",
        handlers=[stdout_handler],
    )



import os
import time

import cv2 as cv

from gamedriver.settings import settings


def tap_xy(x: int, y: int) -> None:
    """Taps a point (x, y).

    Calls the user-provided :py:attr:`settings.Settings.tap_xy` function.

    Args:
        x (int)
        y (int)
    """
    settings["tap_xy"](x, y)


def swipe(x1: int, y1: int, x2: int, y2: int, duration_ms=100) -> None:
    """Swipes from point (x1, y1) to (x2, y2).

    Calls the user-provided :py:attr:`settings.Settings.swipe` function.

    Args:
        x1 (int)
        y1 (int)
        x2 (int)
        y2 (int)
        duration_ms (int, optional): Duration of the swipe in milliseconds.
            Defaults to 100.
    """
    settings["swipe"](x1, y1, x2, y2, duration_ms)


def wait(seconds=1.0) -> None:
    """Sleeps for a time, taking wait adjustments into account.

    Waits can be adjusted by :py:attr:`settings.Settings.wait_scale` and
    :py:attr:`settings.Settings.wait_offset`.

    Args:
        seconds (float, optional): Base time to sleep for. Defaults to 1.
    """
    time.sleep(settings["wait_scale"] * seconds + settings["wait_offset"])


def get_screen(*, grayscale=False) -> cv.typing.MatLike:
    """Gets the device screen.

    Calls the user-provided :py:attr:`settings.Settings.get_screen` function.
    Note that screens should be in the BGR color space.

    That function is augmented by specifying whether there should be a
    conversion to grayscale. This is for functions with `is_grayscale`
    parameters that if set expect argument images to be grayscale.

    Args:
        grayscale (bool, optional): Whether to return the screen in grayscale.
            Defaults to False.

    Returns:
        cv.typing.MatLike: Array representing the screen [height, width,
            channel]

    Examples:
        Save a screenshot::

            cv.imwrite("fname.png", cv.cvtColor(gd.get_screen(), cv.COLOR_BGR2RGB))

        Use grayscale template matching with an already-grayscale template,
        and a custom full path to the template::

            box = gd.match_template(
                gd.get_screen(grayscale=True),
                "my/path/buttons/confirm-gray.png",
                is_grayscale=True,
            )
    """
    screen = settings["get_screen"]()
    if grayscale:
        screen = cv.cvtColor(screen, cv.COLOR_BGR2GRAY)
    return screen


def get_pixel(x: int, y: int) -> tuple[int, int, int]:
    """Gets a pixel from the device screen.

    The pixel should be in color. Note that it should be in the BGR color
    space.

    Args:
        x (int)
        y (int)

    Returns:
        tuple[int, int, int]: Pixel in the BGR color space

    Examples:
        ::

            # Check whether a pixel at (10, 20) is red enough
            px = gd.get_pixel(10, 20)
            is_red = px[0] < 25 and px[1] < 25 and px[2] > 225
    """
    return get_screen()[y, x]


def get_img_path(rel_path: str) -> str:
    """Gets the path to an image.

    The argument path is relative to the base directory
    :py:attr:`settings.Settings.img_path`. Argument images should be without a
    file extension, and will be given the extension
    :py:attr:`settings.Settings.img_ext`.

    Args:
        rel_path (str): Path to the image from the base directory, without a
            file extension

    Returns:
        str: Full path to the image

    Examples
        ::

            # If `img_path` is "/home/user/projects/project/project/img" and
            # `img_ext` is ".png"
            gd.get_img_path("buttons/confirm")
            # "/home/user/projects/project/project/img/buttons/confirm.png"
    """
    ext = settings["img_ext"]
    if not ext.startswith("."):
        ext = f".{ext}"
    return os.path.join(settings["img_path"], f"{rel_path}{ext}")
